/*
 License (MIT License)

 Copyright (c) 2016 Divvil - info@divvil.com

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated  documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 of the Software, and to permit persons to whom the Software is furnished to do
 so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/

var mouseWheelSlider = $.widget('ui.mousewheelslider', {
 version: '1.0.1',
 options: {},
 _create: function() {
  var that = this;
  this.element.bind('wheel', function(e) {
   var change = Math.sign(e.originalEvent.deltaY)*that.element.slider('option', 'step');
   that.element.slider('option', 'value', (that.element.slider('option', 'value')-change));
   e.preventDefault();
  });
 }
});
var colorButton = $.widget('ui.colorbutton', {
 version: '1.0.1',
 options: {
  autofocus: true,
  color: '#40C040',
  dialogHeight: 400,
  dialogTitle: null,
  dialogWidth: 480,
  enableGL: true,
  enableMouseWheel: true,
  innerRadius: 0.15,
  maxDiameter: 2048,
  maxRenderTime: 50
 },
 _create: function() {
  var that = this;
  this.color = this.options.color;
  this.element.button({label: this.color});
  this.element.on('click', function() {
   if (that._wheel) {
    that._wheel.colorwheel('color', that.color);
    that._wheel.dialog('open');
   } else {
    that._wheel = $('<div>');
    that._wheel.dialog({
     title: that.options.dialogTitle,
     minWidth: 280,
     minHeight: 200,
     height: that.options.dialogHeight,
     width: that.options.dialogWidth,
     resize: function() {
      that._wheel.colorwheel('resize');
     },
     close: function(e,ui) {
      that.color = that._wheel.colorwheel('color');
      that.element.find('span').text(that.color);
     }
    });
    that._wheel.colorwheel(that.options);
   }
  });
 },
 color: function(color) {
  if (color) this.color = color;
  else return this.color;
 }
});
var colorWheel = $.widget('ui.colorwheel', {
 version: '1.0.5',
 options: {
  autofocus: false,
  color: '#40C040',
  enableGL: true,
  enableMouseWheel: true,
  innerRadius: 0.15,
  maxDiameter: 2048,
  maxRenderTime: 50
 },
 _create: function() {
  if (this._wheel) return;
  this._detectWorkarounds();
  this.element.addClass('ui-colorwheel');
  this._makeProportional();
  this._outerRadius = 0.48;
  this._makeWheel();
  this._makeSliders();
  this.resize();
  this.color(this.options.color);
 },
 _detectWorkarounds: function() {
  this._supportsGL = true;
  var ua = window.navigator.userAgent.toLowerCase();
  if (!!window.chrome) this._forceRelayoutOnResize = true;
  if (ua.indexOf('trident/') >= 0) this._forceRelayoutOnResize = true;
  if (ua.indexOf('safari') >= 0 &&
      !window.chrome) this._supportsGL = false;
 },
 _makeProportional: function() {
  this._container = $('<div>', {class: 'ui-colorwheel'});
  var left = $('<div>', {class: 'ui-colorwheel-left'});
  var widthLimit = $('<div>', {class: 'ui-colorwheel-widthlimit'});
  var spacer = $('<canvas>', {class: 'ui-colorwheel-spacer'});
  spacer.prop({width: this.options.maxDiameter, height: this.options.maxDiameter});
  this._wheel = $('<div>', {class: 'ui-colorwheel-wheel'});
  widthLimit.append(spacer);
  widthLimit.append(this._wheel);
  left.append(widthLimit);
  this._container.append(left);
  this.element.append(this._container);
 },
 resize: function() {
  if (this._forceRelayoutOnResize) {
   // Workaround for Chrome bug:
   //  https://bugs.webkit.org/show_bug.cgi?id=104872
   //  http://stackoverflow.com/a/3485654
   // Same bug also exists in IE
   this._container[0].style.display = 'none';
   this._container[0].offsetHeight;
   this._container[0].style.display='block';
  }
  this._renderWheel();
 },
 color: function(color) {
  if (color) this._setRGB(this._stringToColor(color));
  else return this._colorToString(this._color);
 },
 _getMouseOffset: function(el, e) {
  var rect = el.getBoundingClientRect();
  return {x: e.clientX-rect.left, y: e.clientY-rect.top};
 },
 _pad: function(s, length, padding) {
  while(s.length < length) s = padding + s;
  return s;
 },
 _stringToColor: function(s) {
  if (s[0]=='#') s = s.substr(1);
  if (s.length != 6) return null;
  r = 1*('0x'+s.substr(0, 2));
  g = 1*('0x'+s.substr(2, 2));
  b = 1*('0x'+s.substr(4, 2));
  if (isNaN(r) || isNaN(g) || isNaN(b)) return undefined;
  return [r, g, b];
 },
 _colorToString: function(color) {
  var s = '#';
  for (var i=0;i<3;i++) s += this._pad(color[i].toString(16), 2, 0);
  return s.toUpperCase();
 },
 _rgb2hsl: function(color, def) {
  if (def && this._hsl2rgb(def).toString() == color.toString()) return def;
  def = def || [0, 0, 0];
  var r=color[0], g=color[1], b=color[2];
  var max = Math.max(r, g, b);
  var min = Math.min(r, g, b);
  var h=def[0], s=def[1];
  var l = Math.round((max+min)/2)/255;
  var d = max - min;
  if (d) {
   var t = max + min;
   s = (l > 0.5) ? d/(2*255 - t) : d/t;
   switch (max) {
    case r:
     h = (g-b)/d + 6;
     break;
    case g:
     h = (b-r)/d + 2;
     break;
    case b:
     h = (r-g)/d + 4;
     break;
   }
   h = (h/6)%1;
  } else if (0<l && l<1) {
   s = 0;
  }
  return [h, s, l];
 },
 _hsl2rgb: function(hsl) {
  var H = 6*(4/3-hsl[0]);
  var rgb = [];
  var l = hsl[2];
  for (var i=0;i<3;i++) {
   var P = (0.5+hsl[1]*(Math.max(-.5, Math.min(.5, 1.5-Math.abs(2-(H+(i<<1))%6)))))*255;
   if (l<0.5) rgb[i] = Math.round(P*2*l);
   else rgb[i] = Math.round(P*(2-2*l) + 255*(2*l-1));
  }
  return rgb;
 },
 _xyl2hsl: function(xyl) {
  var dim = this._wheel.width();
  var radius = dim*this._outerRadius;
  var innerRadius = dim*this.options.innerRadius;
  var c = dim/2;
  var dx = xyl[0]-c;
  var dy = xyl[1]-c;
  var r = Math.sqrt(dx*dx+dy*dy);
  var s = Math.max(0, Math.min(1, (r-innerRadius)/(radius-innerRadius)));
  var h = (Math.atan2(dx, dy)/(2*Math.PI)+.75)%1;
  return [h, s, xyl[2]];
 },
 _hsl2xyl: function(hsl) {
  var dim = this._wheel.width();
  var radius = dim*this._outerRadius;
  var ir = dim*this.options.innerRadius;
  var s = hsl[1];
  var r = ir + s*(radius-ir);
  var h = hsl[0] * Math.PI*2;
  var x = dim/2 + r * Math.cos(h);
  var y = dim/2 - r * Math.sin(h);
  return [x, y, hsl[2]];
 },
 _setHSL: function(value, i) {
  var before = this._hsl.toString();
  if (i>=0) this._hsl[i] = value;
  else this._hsl = value;
  if (before == this._hsl.toString()) return;
  this._hsl[2] = Math.round(255*this._hsl[2])/255;
  this._color = this._hsl2rgb(this._hsl);
  this._updateColor();
 },
 _setRGB: function(value, i) {
  if (value === undefined) return;
  var before = this._color.toString();
  if (i>=0) this._color[i] = value;
  else this._color = value;
  var cs = this._colorToString(this._color);
  if (this._colorstring.val().toUpperCase() != cs) this._colorstring.val(cs);
  this._colorstring.removeClass('ui-colorwheel-invalid');
  if (before == this._color.toString()) return;
  this._hsl = this._rgb2hsl(this._color, this._hsl);
  this._updateColor();
 },
 _makeSlider: function() {
  var canvas = document.createElement('canvas');
  canvas.height = 256;
  canvas.width = 1;
  return $('<div>').append(canvas);
 },
 _makeSliders: function() {
  var that = this;
  this._controls = $('<div>', {class: 'ui-colorwheel-controls'});
  var sliders = $('<div>', {class: 'ui-colorwheel-sliders'});
  var light = this._makeSlider();
  light.slider({
   orientation: 'vertical',
   min: 0,
   max: 255,
   slide: function(e, ui) {
    that._setHSL(ui.value/255, 2);
   },
   change:function(e, ui) {
    that._setHSL(ui.value/255, 2);
   }
  });
  if (this.options.enableMouseWheel) light.mousewheelslider();
  sliders.append(light);
  this._sliders = [light];
  for (var i=0;i<3;i++) {
   var t = this._makeSlider();
   (function(c) {
    t.slider({
     orientation: 'vertical',
     min: 0,
     max: 255,
     slide: function(e, ui) {
      that._setRGB(ui.value, c);
     },
     change: function(e, ui) {
      that._setRGB(ui.value, c);
     }
    });
   })(i);
   if (this.options.enableMouseWheel) t.mousewheelslider();
   sliders.append(t);
   this._sliders.push(t);
  }
  this._colorstring = $('<input>',{autofocus: this.options.autofocus, spellcheck: false, keyup: function(e) {
   var color = that._stringToColor(e.target.value);
   if (color === null) {
    $(e.target).addClass('ui-colorwheel-invalid');
   } else {
    that._setRGB(color);
   }
  }});
  this._controls.append(sliders);
  this._controls.append(this._colorstring);
  this._container.append(this._controls);
 },
 _makeWheel: function() {
  var that = this;
  this._color = this._hsl = [0, 0, 0];
  this._lwheelColor = '#FFF';
  this._wheel.css('opacity', 0);
  var dim = this._wheel.width();
  this._hswheel = $('<div>');
  this._lwheel = $('<div>');
  this._field = $('<div>', {class: 'ui-colorwheel-field'});
  this._wheel.append(this._field);
  this._wheel.append(this._hswheel);
  this._wheel.append(this._lwheel);
  this._marker = $('<div>', {class: 'ui-colorwheel-marker'});
  this._wheel.append(this._marker);
  this._wheel.append($('<svg><circle class="ui-colorwheel-hoverarea" cx="50%" cy="50%" r="48%" fill-opacity="0"/><circle cx="50%" cy="50%" r="' + (100*this.options.innerRadius) + '%" fill-opacity="0"/></svg>'));
  var circle = this._wheel.find('.ui-colorwheel-hoverarea');
  circle.on('mousedown', function(e) {
   e.preventDefault();
   var mc = that._getMouseOffset(that._wheel[0], e);
   that._setHSL(that._xyl2hsl([mc.x, mc.y, that._hsl[2]]));
   that.element.addClass('ui-colorwheel-nodrop');
   $(document).on('mousemove.ui-colorwheel', function(e) {
    var mc = that._getMouseOffset(that._wheel[0], e);
    that._setHSL(that._xyl2hsl([mc.x, mc.y, that._hsl[2]]));
   });
   $(document).one('mouseup', function() {
    $(document).off('mousemove.ui-colorwheel');
    that.element.removeClass('ui-colorwheel-nodrop');
   });
  });
  if (this.options.enableMouseWheel) {
   circle.on('wheel', function(e) {
    if (e.originalEvent.deltaY) {
     var s = Math.min(1, Math.max(0, that._hsl[1]-Math.sign(e.originalEvent.deltaY)/256));
     that._setHSL(s, 1);
    }
    if (e.originalEvent.deltaX) {
     that._setHSL((1+that._hsl[0]+Math.sign(e.originalEvent.deltaX)/256) % 1, 0);
    }
    e.preventDefault();
   });
  }
 },
 _updateColor: function() {
  var that = this;
  var dim = this._wheel.width();
  var xyl = this._hsl2xyl(this._hsl);
  this._marker.css('left', (100*xyl[0]/dim)+'%');
  this._marker.css('top', (100*xyl[1]/dim)+'%');
  if (this._updatePending) return;
  this._updatePending = setTimeout(function() {
   var xyl = that._hsl2xyl(that._hsl);
   for (var j=0;j<3;j++) {
    that._sliders[j+1].slider('value', that._color[j]);
   }
   that._updateSliderBackgrounds();
   that._sliders[0].slider('value', Math.round(that._hsl[2]*255));
   $(that._lwheel).css('opacity', Math.abs(2*xyl[2]-1));
   var lwheelColor = (xyl[2]<0.5) ? '#000' : '#FFF';
   if (lwheelColor!=that._lwheelColor) that._updateLWheel(lwheelColor);
   var cs = that._colorToString(that._color);
   that._field.css('background', cs);
   var event = $.Event();
   if (that._initialized) that._trigger('change', event, {color: cs});
   that._initialized = true;
   delete that._updatePending;
  }, 10);
 },
 _updateLWheel: function(color) {
  if (this._lwheel.getContext) {
   this._lwheelColor = color;
   var ctx = this._lwheel.getContext('2d');
   ctx.globalCompositeOperation = 'source-atop';
   ctx.fillStyle = color;
   ctx.fillRect(0, 0,this._lwheel.width, this._lwheel.height);
   $(this._lwheel).css('opacity', Math.abs(2*this._hsl[2]-1));
  }
 },
 _updateSliderBackgrounds: function() {
  var ctx = this._sliders[0].find('canvas')[0].getContext('2d');
  var gradient = ctx.createLinearGradient(0, 0, 0, 256);
  var hsl = this._hsl.slice();
  hsl[2] = 0.5;
  gradient.addColorStop(0, '#FFFFFF');
  gradient.addColorStop(0.5, this._colorToString(this._hsl2rgb(hsl)));
  gradient.addColorStop(1, '#000000');
  ctx.fillStyle=gradient;
  ctx.fillRect(0, 0, 1, 256);
  for (var i=0;i<3;i++) {
   var rgb = this._color.slice();
   gradient = ctx.createLinearGradient(0, 0, 0, 256);
   rgb[i] = 255;
   gradient.addColorStop(0, this._colorToString(rgb));
   rgb[i] = 0;
   gradient.addColorStop(1, this._colorToString(rgb));
   ctx = this._sliders[i+1].find('canvas')[0].getContext('2d');
   ctx.fillStyle = gradient;
   ctx.fillRect(0, 0, 1, 256);
  }
 },
 _renderWheel: function() {
  var start = Date.now();
  var dim = Math.floor(this._wheel.width());
  var rip = this._renderInProgress;
  if (!rip || dim > rip.dim) {
   if (dim <= this._wheelDim) return;
   this._renderTimeout = clearTimeout(this._renderTimeout);
   this._wheelDim = dim;
   rip = {
    start: start,
    lastRender: start,
    iterations: 1,
    dim: dim,
    y: 0,
    outerRadius: dim * this._outerRadius,
    wheelEdge: 2.0,
    innerRadius: dim * this.options.innerRadius,
    hswheel: document.createElement('canvas')
   };
   rip.hswheel.width = rip.hswheel.height = dim;
   try {
    rip.gl = this._supportsGL && this.options.enableGL && rip.hswheel.getContext('webgl', {premultipliedAlpha: false});
   } catch (e) {
   }
   if (!rip.gl) {
    rip.ctx = rip.hswheel.getContext('2d');
    rip.img = rip.ctx.getImageData(0, 0, dim, dim);
   }
   this._renderInProgress = rip;
   return this._asynchRender();
  }
  var hswheel;
  if (rip.gl) hswheel = this._renderHSWheelGL(rip);
  else hswheel = this._renderHSWheel2d(rip);
  if (hswheel) {
   $(this._hswheel).replaceWith(hswheel);
   this._hswheel = hswheel;
   delete this._renderInProgress;
   var lwheel = document.createElement('canvas');
   var ctx = lwheel.getContext('2d');
   lwheel.width = hswheel.width;
   lwheel.height = hswheel.height;
   ctx.drawImage(hswheel, 0, 0);
   $(this._lwheel).replaceWith(lwheel);
   this._lwheel = lwheel;
   this._updateLWheel(this._lwheelColor);
   this._wheel.css('opacity', 1);
  }
 },
 _asynchRender: function() {
  var that = this;
  this._renderTimeout = setTimeout(function() {
   that._renderWheel();
  }, 10);
 },
 _renderHSWheel2d: function(rip) {
  var that = this;
  var c = rip.dim/2;
  var a, s;
  for (var y=rip.y;y<rip.dim;y++) {
   var now = Date.now();
   if (now - rip.lastRender > this.options.maxRenderTime) {
    rip.lastRender = now;
    rip.iterations++;
    rip.y = y;
    return this._asynchRender();
   }
   for (var x=0;x<rip.dim;x++) {
    var dx = x-c;
    var dy = y-c;
    var r = Math.sqrt(dx*dx+dy*dy);
    if (r < rip.innerRadius) {
     s = 0;
     a = Math.max(0, 1-(rip.innerRadius-r)/rip.wheelEdge);
    } else if (r <= rip.outerRadius) {
     s = (r-rip.innerRadius)/(rip.outerRadius-rip.innerRadius);
     a = 1;
    } else {
     s = 1;
     a = Math.max(0, 1-(r-rip.outerRadius)/(rip.wheelEdge-1));
    }
    var j = (y*rip.dim+x)<<2;
    if (rip.innerRadius-rip.wheelEdge < r && r < rip.outerRadius+rip.wheelEdge) {
     var h = 3*Math.atan2(-dx, dy)/Math.PI + 3.5;
     for (var i=0;i<3;i++) {
      var P = .5+s*(Math.max(-.5, Math.min(.5, 1.5-Math.abs(2-(h+(i<<1))%6))));
      rip.img.data[j+i] = P*256;
     }
     rip.img.data[j+3] = a*255;
    } else {
     rip.img.data[j+3] = 0;
    }
   }
  }
  rip.ctx.putImageData(rip.img, 0, 0);
  return rip.hswheel;
 },
 _renderHSWheelGL: function(rip) {
  var gl = rip.gl;
  var vs = [
   'attribute vec2 position;',
   'void main() {',
   'gl_Position = vec4(position, 0, 1);',
   '}'
  ].join('\n');
  var fs = [
   'precision mediump float;',
   'uniform vec4 dims;',
   '#define PI 3.141592653',
   'void main() {',
   ' vec2 d, e;',
   ' float r, s, a;',
   ' d = (gl_FragCoord.xy-dims.xx);',
   ' e = d * d;',
   ' r = sqrt(e.x+e.y);',
   ' s = clamp((r-dims.z)/(dims.y-dims.z), 0.0, 1.0);',
   ' a = max(r-dims.y, dims.z-r);',
   ' a = clamp(1.0-(a/dims.w), 0.0, 1.0);',
   ' gl_FragColor.a = a;',
   ' if (a > 0.0) {',
   '  r = 0.5+3.0*atan(d.x, d.y)/PI;',
   '  for(int i=0;i<3;i++) {',
   '   gl_FragColor[i] = 0.5+s*clamp(1.5-abs(2.0-mod(r+float(2*i), 6.0)), -.5, .5);',
   '  }',
   ' }',
   '}'
  ].join('\n');

  gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);
  function createShader(type, code) {
    var shader = gl.createShader(type);
    gl.shaderSource(shader, code);
    gl.compileShader(shader);
    return shader;
  }
  var vertexShader = createShader(gl.VERTEX_SHADER, vs);
  var fragmentShader = createShader(gl.FRAGMENT_SHADER, fs);
  var program = gl.createProgram();
  gl.attachShader(program, vertexShader);
  gl.attachShader(program, fragmentShader);
  gl.linkProgram(program);
  gl.useProgram(program);

  var buffer = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
  gl.bufferData(
   gl.ARRAY_BUFFER,
   new Float32Array([
    -1,-1, 1,-1,-1, 1,
    -1, 1, 1,-1, 1, 1
   ]),
   gl.STATIC_DRAW
  );
  var position = gl.getAttribLocation(program, 'position');
  gl.enableVertexAttribArray(position);
  gl.vertexAttribPointer(position, 2, gl.FLOAT, false, 0, 0);
  var dims = gl.getUniformLocation(program, 'dims');
  gl.uniform4f(dims, gl.drawingBufferWidth/2, rip.outerRadius, rip.innerRadius, rip.wheelEdge);
  gl.drawArrays(gl.TRIANGLES, 0, 6);
  return rip.hswheel;
 }
});
